from __future__ import annotations

import string
from collections import ChainMap
from math import hypot
from pathlib import Path
from typing import Optional

import plotly.express as px
import procrunner
import workflows.recipe
from pydantic import BaseModel, Field
from pydantic.error_wrappers import ValidationError
from workflows.services.common_service import CommonService


class MotionCorrParameters(BaseModel):
    pix_size: float
    ctf: dict
    movie: str = Field(..., min_length=1)
    mrc_out: str = Field(..., min_length=1)
    patch_size: int = 5
    gpu: int = 0
    gain_ref: str = None
    mc_uuid: int
    rot_gain: int = None
    flip_gain: int = None
    dark: str = None
    use_gpus: int = None
    sum_range: Optional[tuple] = None
    iter: int = None
    tol: float = None
    throw: int = None
    trunc: int = None
    fm_ref: int = None
    kv: int = None
    fm_dose: float = None
    fm_int_file: str = None
    mag: Optional[tuple] = None
    ft_bin: float = None
    serial: int = None
    in_suffix: str = None
    eer_sampling: int = None
    out_stack: int = None
    bft: Optional[tuple] = None
    group: int = None
    detect_file: str = None
    arc_dir: str = None
    in_fm_motion: int = None
    split_sum: int = None
    movie_id: int

    class Config:
        ignore_extra = True


class ChainMapWithReplacement(ChainMap):
    def __init__(self, *maps, substitutions=None) -> None:
        super().__init__(*maps)
        self._substitutions = substitutions

    def __getitem__(self, k):
        v = super().__getitem__(k)
        if self._substitutions and isinstance(v, str) and "$" in v:
            template = string.Template(v)
            return template.substitute(**self._substitutions)
        return v


class MotionCorr(CommonService):
    """
    A service for motion correcting cryoEM movies using MotionCor2
    """

    # Human readable service name
    _service_name = "DLS MotionCorr"

    # Logger name
    _logger_name = "relion.zocalo.motioncorr"

    # Values to extract for ISPyB
    shift_list = []

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Motion correction service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "motioncorr",
            self.motion_correction,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_mc_output(self, line: str):
        if not line:
            return

        if line.startswith("...... Frame"):
            line_split = line.split()
            self.shift_list.append((float(line_split[-2]), float(line_split[-1])))

    def motion_correction(self, rw, header: dict, message: dict):
        class RW_mock:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = RW_mock()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"], "output": None}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        command = ["MotionCor2"]

        parameter_map = ChainMapWithReplacement(
            message if isinstance(message, dict) else {},
            rw.recipe_step["parameters"],
            substitutions=rw.environment,
        )

        try:
            if isinstance(message, dict):
                mc_params = MotionCorrParameters(**{**dict(parameter_map), **message})
            else:
                mc_params = MotionCorrParameters(**{**dict(parameter_map)})
        except (ValidationError, TypeError):
            self.log.warning(
                f"Motion correction parameter validation failed for message: {message} and recipe parameters: {rw.recipe_step.get('parameters', {})}"
            )
            rw.transport.nack(header)
            return
        if Path(mc_params.mrc_out).is_file():
            self.log.info(f"File exists {mc_params.mrc_out}")
            rw.transport.ack(header)
            return
        movie = mc_params.movie
        if movie.endswith(".mrc"):
            input_flag = "-InMrc"
        elif movie.endswith((".tif", ".tiff")):
            input_flag = "-InTiff"
        elif movie.endswith(".eer"):
            input_flag = "-InEer"
        else:
            self.log.error(f"No input flag found for movie {movie}")
            input_flag = None
            rw.transport.nack(header)
        command.extend([input_flag, movie])

        mc_flags = {
            "mrc_out": "-OutMrc",
            "patch": "-Patch",
            "pix_size": "-PixSize",
            "gain_ref": "-Gain",
            "rot_gain": "-RotGain",
            "flip_gain": "-FlipGain",
            "dark": "-Dark",
            "gpu": "-Gpu",
            "use_gpus": "-UseGpus",
            "sum_range": "-SumRange",
            "iter": "-Iter",
            "tol": "-Tol",
            "throw": "-Throw",
            "trunc": "-Trunc",
            "fm_ref": "-FmRef",
            "kv": "-Kv",
            "fm_dose": "-FmDose",
            "fm_int_file": "-FmIntFile",
            "mag": "-Mag",
            "ft_bin": "-FtBin",
            "serial": "-Serial",
            "in_suffix": "-InSuffix",
            "eer_sampling": "-EerSampling",
            "out_stack": "-OutStack",
            "bft": "-Bft",
            "group": "-Group",
            "detect_file": "-DetectFile",
            "arc_dir": "-ArcDir",
            "in_fm_motion": "-InFmMotion",
            "split_sum": "-SplitSum",
        }

        for k, v in mc_params.dict().items():
            if v and (k in mc_flags):
                if type(v) is tuple:
                    command.extend((mc_flags[k], " ".join(str(_) for _ in v)))
                else:
                    command.extend((mc_flags[k], str(v)))

        self.log.info(f"Input: {movie} Output: {mc_params.mrc_out}")

        result = procrunner.run(command=command, callback_stdout=self.parse_mc_output)
        if result.returncode:
            self.log.error(
                f"Motion correction of {movie} failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        # Forward results to ctffind
        self.log.info("Sending to ctf")
        mc_params.ctf["input_image"] = mc_params.mrc_out
        mc_params.ctf["mc_uuid"] = mc_params.mc_uuid
        mc_params.ctf["pix_size"] = mc_params.pix_size
        if isinstance(rw, RW_mock):
            rw.transport.send(  # type: ignore
                destination="ctffind",
                message={"parameters": mc_params.ctf, "content": "dummy"},
            )
        else:
            rw.send_to("ctf", mc_params.ctf)

        # Extract results for ispyb
        total_x_shift = sum([item[0] for item in self.shift_list])
        total_y_shift = sum([item[1] for item in self.shift_list])
        total_motion = hypot(total_x_shift, total_y_shift)
        each_total_motion = [hypot(item[0], item[1]) for item in self.shift_list]
        average_motion_per_frame = sum(each_total_motion) / len(self.shift_list)

        drift_plot_x = [range(0, len(self.shift_list))]
        drift_plot_y = each_total_motion
        fig = px.scatter(x=drift_plot_x, y=drift_plot_y)
        drift_plot_name = str(Path(mc_params.movie).stem) + "_drift_plot.json"
        plot_path = Path(mc_params.mrc_out).parent / drift_plot_name
        snapshot_path = Path(mc_params.mrc_out).with_suffix(".jpeg")
        fig.write_json(plot_path)

        ispyb_parameters = {
            "first_frame": 1,
            "last_frame": len(self.shift_list),
            "total_motion": total_motion,
            "average_motion_per_frame": average_motion_per_frame,
            "drift_plot_full_path": str(plot_path),
            "micrograph_snapshot_full_path": str(snapshot_path),
            "micrograph_full_path": str(mc_params.mrc_out),
            "patches_used_x": mc_params.patch_size,
            "patches_used_y": mc_params.patch_size,
            "buffer_store": mc_params.mc_uuid,
            "dose_per_frame": mc_params.fm_dose,
        }

        # Forward results to ISPyB
        ispyb_parameters.update(
            {
                "ispyb_command": "buffer",
                "buffer_command": {"ispyb_command": "insert_motion_correction"},
            }
        )
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, RW_mock):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb", ispyb_parameters)

        # Forward results to murfey
        self.log.info("Sending to Murfey")
        if isinstance(rw, RW_mock):
            rw.transport.send(
                "murfey_feedback",
                {
                    "register": "motion_corrected",
                    "movie": mc_params.movie,
                    "mrc_out": mc_params.mrc_out,
                    "movie_id": mc_params.movie_id,
                },
            )
        else:
            rw.send_to(
                "murfey",
                {
                    "register": "motion_corrected",
                    "movie": mc_params.movie,
                    "mrc_out": mc_params.mrc_out,
                    "movie_id": mc_params.movie_id,
                },
            )

        # Forward results to images service
        self.log.info(f"Sending to images service {mc_params.mrc_out}")
        if isinstance(rw, RW_mock):
            rw.transport.send(
                destination="images",
                message={
                    "parameters": {"images_command": "mrc_to_jpeg"},
                    "file": mc_params.mrc_out,
                },
            )
        else:
            rw.send_to(
                "images",
                {
                    "parameters": {"images_command": "mrc_to_jpeg"},
                    "file": mc_params.mrc_out,
                },
            )

        rw.transport.ack(header)
        self.shift_list = []



from __future__ import annotations

import os
import pathlib
import warnings
from collections import Counter
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from threading import RLock
from typing import Optional, Tuple

from gemmi import cif

try:
    from graphviz import Digraph
except ImportError:
    pass
import calendar
import copy
import datetime

from relion._parser.processgraph import ProcessGraph
from relion._parser.processnode import ProcessNode


def _max_class(ini_model_dir: Path) -> int:
    number_list = [entry.stem[6:9] for entry in (ini_model_dir).glob("run_it*.star")]
    last_iteration_number = max(
        (int(n) for n in number_list if n.isnumeric()), default=0
    )
    data_file = ini_model_dir / f"run_it{last_iteration_number:03d}_data.star"
    gemmi_readable_path = os.fspath(data_file)
    star_doc = cif.read_file(gemmi_readable_path)
    class_number_values = star_doc[1].find_loop("_rlnClassNumber")
    counts = Counter(list(class_number_values))
    return int(counts.most_common(1)[0][0])


class RelionPipeline:
    def __init__(self, origin, graphin=ProcessGraph("nodes", []), locklist=None):
        self.origin = origin
        self._nodes = graphin
        self._connected = {}
        self.origins = {}
        self._job_nodes = ProcessGraph("job nodes", [])
        self._jobtype_nodes = ProcessGraph("job type nodes", [])
        self._connected_jobs = {}
        self.job_origins = {}
        self._jobs_collapsed = False
        self.locklist = locklist or []
        self.preprocess = []

    def __iter__(self):
        if not self._jobs_collapsed:
            self._collapse_jobs_to_jobtypes()
        return iter(self._jobtype_nodes._node_list)

    @property
    def _plock(self):
        return DummyLock()

    def _star_doc(self, star_path):
        gemmi_readable_path = os.fspath(star_path)
        if star_path in self.locklist:
            with self._plock as pl:
                if pl.obtained:
                    star_doc = cif.read_file(gemmi_readable_path)
                else:
                    # effectively return an empty star file
                    star_doc = cif.Document()
            return star_doc
        return cif.read_file(gemmi_readable_path)

    def _request_star_values(self, star_doc, column, search=None):
        if search is None:
            search = column
        block_number = None
        for block_index, block in enumerate(star_doc):
            if list(block.find_loop(search)):
                block_number = block_index
                break
        else:
            return []
        data_block = star_doc[block_number]
        values = data_block.find_loop(column)
        return list(values)

    def _load_file_nodes_from_star(self, star_doc):
        return ProcessGraph(
            "file nodes",
            [
                ProcessNode(pathlib.Path(p))
                for p in self._request_star_values(star_doc, "_rlnPipeLineNodeName")
            ],
        )

    def _load_job_nodes_from_star(self, star_doc):
        drops = {}
        drops["InitialModel"] = ["batch_number"]
        return ProcessGraph(
            "job nodes",
            [
                ProcessNode(
                    pathlib.Path(p),
                    independent=True,
                    alias=al,
                    drop=drops.get(p.split("/")[0]),
                )
                for p, al in zip(
                    self._request_star_values(star_doc, "_rlnPipeLineProcessName"),
                    self._request_star_values(star_doc, "_rlnPipeLineProcessAlias"),
                )
            ],
        )

    def load_nodes_from_star(self, star_path):
        self._nodes.wipe()
        star_doc_from_path = self._star_doc(star_path)
        file_nodes = self._load_file_nodes_from_star(star_doc_from_path)
        job_nodes = self._load_job_nodes_from_star(star_doc_from_path)
        self._nodes.extend(file_nodes)
        self._nodes.extend(job_nodes)
        binding_pairs = [
            (ProcessNode(p1), ProcessNode(p2))
            for p1, p2 in zip(
                self._request_star_values(
                    star_doc_from_path, "_rlnPipeLineEdgeFromNode"
                ),
                self._request_star_values(
                    star_doc_from_path,
                    "_rlnPipeLineEdgeProcess",
                    search="_rlnPipeLineEdgeFromNode",
                ),
            )
        ]
        binding_pairs.extend(
            [
                (ProcessNode(p1), ProcessNode(p2))
                for p1, p2 in zip(
                    self._request_star_values(
                        star_doc_from_path,
                        "_rlnPipeLineEdgeProcess",
                        search="_rlnPipeLineEdgeToNode",
                    ),
                    self._request_star_values(
                        star_doc_from_path, "_rlnPipeLineEdgeToNode"
                    ),
                )
            ]
        )
        for f, t in binding_pairs:
            self._nodes[self._nodes.index(f._path)].link_to(
                self._nodes[self._nodes.index(t._path)]
            )
            if str(f._path.parent.parent) == "Select" and f._path.name.startswith(
                "particles_split"
            ):
                self._nodes[self._nodes.index(f._path)].environment[
                    "batch_number"
                ] = f._path.stem.replace("particles_split", "")
                self._nodes[self._nodes.index(f._path)].environment["inject"] = [
                    ("batch_number", "batch_number")
                ]
                self._nodes[self._nodes.index(f._path)].propagate(
                    ("batch_number", "batch_number")
                )
                self._nodes[self._nodes.index(f._path)].propagate(("inject", "inject"))
            if str(f._path.parent.parent) == "InitialModel" and "class" in f._path.name:
                self._nodes[self._nodes.index(f._path)].environment[
                    "init_model_class_num"
                ] = int(f._path.stem.split("class")[-1].split("_")[0])
                self._nodes[self._nodes.index(f._path)].propagate(
                    ("init_model_class_num", "init_model_class_num")
                )
            elif (
                str(f._path.parent.parent) == "InitialModel"
                and "initial" in f._path.name
            ):
                self._nodes[self._nodes.index(f._path)].environment[
                    "init_model_class_num"
                ] = _max_class(Path(star_path).parent / f._path.parent)
                self._nodes[self._nodes.index(f._path)].propagate(
                    ("init_model_class_num", "init_model_class_num")
                )
        self._nodes._split_connected(self._connected, self.origin, self.origins)
        self._set_job_nodes(star_doc_from_path)

    def check_job_node_statuses(self, basepath):
        for node in self._job_nodes:
            success = basepath / node._path / "RELION_JOB_EXIT_SUCCESS"
            success_pipeliner = basepath / node._path / "PIPELINER_JOB_EXIT_SUCCESS"
            failure = basepath / node._path / "RELION_JOB_EXIT_FAILURE"
            aborted = basepath / node._path / "RELION_JOB_EXIT_ABORTED"
            # the try/except blocks below catch the case where Relion removes
            # the SUCCESS/FAILURE file in between checking for its existence
            # and checking its modification time
            try:
                node.environment["end_time_stamp"] = datetime.datetime.fromtimestamp(
                    failure.stat().st_mtime
                )
                node.environment["status"] = False
                continue
            except FileNotFoundError:
                pass
            try:
                node.environment["end_time_stamp"] = datetime.datetime.fromtimestamp(
                    success.stat().st_mtime
                )
                node.environment["status"] = True
                continue
            except FileNotFoundError:
                pass
            try:
                node.environment["end_time_stamp"] = datetime.datetime.fromtimestamp(
                    success_pipeliner.stat().st_mtime
                )
                node.environment["status"] = True
                continue
            except FileNotFoundError:
                pass
            try:
                node.environment["end_time_stamp"] = datetime.datetime.fromtimestamp(
                    aborted.stat().st_mtime
                )
                node.environment["status"] = False
                continue
            except FileNotFoundError:
                pass
            node.environment["status"] = None

    def _set_job_nodes(self, star_doc):
        self._job_nodes = copy.deepcopy(self._nodes)
        file_nodes = self._load_file_nodes_from_star(star_doc)
        for fnode in file_nodes:
            if str(
                fnode._path.parent.parent
            ) == "Select" and fnode._path.name.startswith("particles_split"):
                self._job_nodes.remove_node(fnode._path, advance=True)
            else:
                self._job_nodes.remove_node(fnode._path)
        self._job_nodes._split_connected(
            self._connected_jobs, self.origin, self.job_origins
        )

    def _collapse_jobs_to_jobtypes(self):
        ordered_graph = []
        if len(self._nodes) == 0:
            self._jobtype_nodes = ProcessGraph("job type nodes", [])
            return
        self._job_nodes.node_explore(
            self._job_nodes[self._job_nodes.index(self.origin)], ordered_graph
        )
        self._jobtype_nodes = ProcessGraph(
            "job type nodes", copy.deepcopy(ordered_graph)
        )
        for node in self._jobtype_nodes:
            node.environment["job"] = node._path.name
            job_string = str(node._path.name)
            node._path = node._path.parent
            for inode in node._in:
                inode._link_traffic[node.nodeid] = inode._link_traffic[node.nodeid]
            node._name = str(node._path)
            if node.name == "InitialModel":
                node.environment["ini_model_job_string"] = job_string
            else:
                node.environment["job_string"] = job_string
        self._jobs_collapsed = True

    def show_all_nodes(self):
        self._nodes.show_all_nodes()

    def show_job_nodes(self, basepath):
        try:
            digraph = Digraph(format="svg")
        except Exception:
            warnings.warn(
                "Failed to create nodes display. Your environment may not have graphviz avaliable."
            )
            return
        digraph.attr(rankdir="LR")
        running_node = self.current_job
        for node in self._job_nodes:
            if node._path == running_node:
                bordercolour = "teal"
            else:
                bordercolour = "purple"
            if node.environment["status"]:
                node_colour = "mediumseagreen"
            elif node.environment["status"] is None:
                node_colour = "lightgray"
            else:
                node_colour = "orangered"
            if node.environment["start_time_stamp"] is not None:
                sstamp = node.environment["start_time_stamp"]
            else:
                sstamp = "???"
            if node.environment["end_time_stamp"] is not None:
                estamp = node.environment["end_time_stamp"]
            else:
                estamp = "???"
            if (
                node.environment.get("alias") is not None
                and node.environment.get("alias") != "None"
            ):
                nodename = node.environment.get("alias")
            else:
                nodename = str(node._path)
            digraph.node(
                name=nodename,
                shape="hexagon",
                style="filled",
                fillcolor=node_colour,
                color=bordercolour,
                tooltip=f"Start: {sstamp}&#013;End: {estamp}&#013;Path: {node._path}",
            )
            for next_node in node:
                stime = next_node.environment["start_time"]
                etime = next_node.environment["end_time"]
                jcount = next_node.environment["job_count"]

                if (
                    next_node.environment["alias"] is not None
                    and next_node.environment["alias"] != "None"
                ):
                    next_nodename = next_node.environment["alias"]
                else:
                    next_nodename = str(next_node._path)

                if stime is not None and etime is not None and jcount != 1:
                    digraph.edge(
                        nodename,
                        next_nodename,
                        label=f"{stime} / {etime} [{jcount}]",
                    )
                elif stime is not None and etime is not None:
                    digraph.edge(
                        nodename,
                        next_nodename,
                        label=f"{stime} / {etime}",
                    )
                elif stime is not None:
                    digraph.edge(
                        nodename,
                        next_nodename,
                        label=f"{stime} / ???",
                    )
                elif etime is not None:
                    digraph.edge(
                        nodename,
                        next_nodename,
                        label=f"??? / {etime}",
                    )
                else:
                    digraph.edge(nodename, next_nodename, label="??? / ???")
        digraph.render(basepath / "Pipeline" / "relion_pipeline_jobs.gv")

    def collect_cluster_info(self, basepath: pathlib.Path):
        # preload schedule log files
        logs = {
            "preproc": self._get_log(basepath / "pipeline_PREPROCESS.log"),
            "class2d": self._get_log(basepath / "pipeline_CLASS2D.log"),
            "inimodel": self._get_log(basepath / "pipeline_INIMODEL.log"),
            "class3d": self._get_log(basepath / "pipeline_CLASS3D.log"),
            "ibgroup": self._get_log(basepath / "pipeline_ICEBREAKER_GROUP.log"),
        }
        with ThreadPoolExecutor(max_workers=10) as pool:
            lock = RLock()
            threads = []
            for job in self._job_nodes:
                sched = ""
                if (
                    str(job._path.parent)
                    in [
                        "Import",
                        "MotionCorr",
                        "CtfFind",
                        "External",
                        "AutoPick",
                        "Select",
                        "Extract",
                    ]
                    and job.environment["alias"]
                    and "Icebreaker_group" not in job.environment["alias"]
                ):
                    sched = "preproc"
                elif str(job._path.parent) == "Class2D":
                    sched = "class2d"
                elif str(job._path.parent) == "InitialModel":
                    sched = "inimodel"
                elif str(job._path.parent) == "Class3D":
                    sched = "class3d"
                elif (
                    job.environment["alias"]
                    and "Icebreaker_group" in job.environment["alias"]
                ):
                    sched = "ibgroup"
                pool_sub = pool.submit(
                    self._job_cluster_info,
                    job,
                    basepath,
                    lock,
                    schedule_log=logs.get(sched),
                )
                threads.append(pool_sub)
            # wait for threads to finish
            [t.result() for t in threads]
            # use the motion correction num micrographs processed counts as
            # the icebreaker counts as those aren't written to the log files
            # needs to be done after all cluster job info is collected
            for job in self._job_nodes:
                if (
                    str(job._path.parent) == "External"
                    and "Icebreaker" in job.environment["alias"]
                ):
                    if (
                        job.environment["cluster_job_ids"]
                        and job._in[0].environment["cluster_job_mic_counts"] is not None
                    ):
                        job.environment["cluster_job_mic_counts"] = job._in[
                            0
                        ].environment["cluster_job_mic_counts"][
                            : len(job.environment["cluster_job_ids"])
                        ]

    def _job_cluster_info(
        self,
        job: str,
        basepath: pathlib.Path,
        lock: RLock,
        schedule_log: Optional[list] = None,
    ):
        try:
            with open(basepath / job._path / "run.out") as logfile:
                log = logfile.readlines()
                (
                    job.environment["cluster_job_ids"],
                    job.environment["cluster_job_start_times"],
                    job.environment["cluster_job_mic_counts"],
                ) = self._parse_out_log(log)
            with open(basepath / job._path / "note.txt") as logfile:
                log = logfile.readlines()
                cmd = None
                for line in log:
                    if "which" in line:
                        cmd = line.split()[1].replace("`", "")
                job.environment["cluster_command"] = cmd
        except FileNotFoundError:
            job.environment["cluster_job_ids"] = []
            job.environment["cluster_job_start_times"] = []
            job.environment["cluster_job_mic_counts"] = []
        if schedule_log:
            with lock:
                job.environment["job_start_times"] = self._get_job_times(
                    schedule_log, job._path
                )

    def _parse_out_log(self, outlog: list) -> Tuple[list]:
        cluster_ids = []
        job_count = 0
        mic_counts = []
        t = []
        for line in outlog:
            if "with job ID" in line:
                cluster_ids.append(line.split()[-1])
                job_count += 1
                mic_counts.append(0)
                # all of this is annoying stuff to deal with Relion writing its progress bar
                # while the cluster info we're interested in is written on the same line of the
                # output log
                time_string = ":".join(line.split(":")[:3])
                time_string = time_string.split(".")[-2]
                for c in ("~", ",", "_", '"', ">", "(", ")", "[", "o", "]"):
                    time_string = time_string.replace(c, "")
                time_string = " ".join(time_string.split(" ")[-2:])
                t.append(datetime.datetime.strptime(time_string, "%Y-%m-%d %H:%M:%S"))
            if (
                ("*" in line or "Filtering" in line)
                and (".mrc" in line or ".tiff" in line)
                and job_count
            ):
                mic_counts[job_count - 1] += 1
        if all(cid.isnumeric() for cid in cluster_ids):
            mic_counts = None
        return cluster_ids, t, mic_counts

    def _get_job_times(self, log: list, job_path: pathlib.Path) -> list:
        times = []
        for lindex, line in enumerate(log):
            if "Executing" in line and str(job_path) in line:
                split_line = log[lindex - 1].split()
                time_split = split_line[4].split(":")
                dtime = datetime.datetime(
                    year=int(split_line[5]),
                    month=list(calendar.month_abbr).index(split_line[2]),
                    day=int(split_line[3]),
                    hour=int(time_split[0]),
                    minute=int(time_split[1]),
                    second=int(time_split[2]),
                )
                times.append(dtime)
        return times

    @staticmethod
    def _get_log(log_path):
        try:
            with open(log_path) as f:
                log = f.readlines()
            return log
        except FileNotFoundError:
            return []

    def collect_job_times(self, schedule_logs, preproc_log=None):
        for job in self._job_nodes:
            jtime, jcount = self._lookup_job_time(schedule_logs, job)
            job.environment["start_time_stamp"] = jtime
            job.environment["job_count"] = jcount
        self._calculate_relative_job_times()
        self.preprocess = self._get_pipeline_jobs(preproc_log)

    def _get_pipeline_jobs(self, logfile):
        if logfile is None:
            return []
        if logfile.is_file():
            with open(logfile, "r") as lfile:
                try:
                    return [
                        self._job_nodes[
                            self._job_nodes.index(
                                pathlib.PurePosixPath(line.split()[1])
                            )
                        ]
                        for line in lfile
                        if line.startswith(" - ")
                    ]
                # if it doesn't find the job in self._job_nodes then return empty list
                # should sort itself out later
                except ValueError:
                    return []
        return []

    def _calculate_relative_job_times(self):
        times = [j.environment["start_time_stamp"] for j in self._job_nodes]
        etimes = [j.environment["end_time_stamp"] for j in self._job_nodes]
        already_started_times = [t for t in times if t is not None]
        try:
            origin = sorted(already_started_times)[0]
        except IndexError:
            return
        relative_times = [(t - origin).total_seconds() for t in times if t is not None]
        just_seconds = lambda tdelta: tdelta - datetime.timedelta(
            microseconds=tdelta.microseconds
        )
        for node, rt in zip(self._job_nodes, relative_times):
            node.environment["start_time"] = just_seconds(
                datetime.timedelta(seconds=rt)
            )
        relative_etimes = [
            (t - origin).total_seconds() for t in etimes if t is not None
        ]
        for node, rt in zip(self._job_nodes, relative_etimes):
            node.environment["end_time"] = just_seconds(datetime.timedelta(seconds=rt))

    def _lookup_job_time(self, schedule_logs, job):
        time = None
        jobcount = 0
        for log in schedule_logs:
            with open(log, "r") as slf:
                lines = slf.readlines()
                for lindex, line in enumerate(lines):
                    if "Executing" in line and str(job._path) in line:
                        split_line = lines[lindex - 1].split()
                        time_split = split_line[4].split(":")
                        dtime = datetime.datetime(
                            year=int(split_line[5]),
                            month=list(calendar.month_abbr).index(split_line[2]),
                            day=int(split_line[3]),
                            hour=int(time_split[0]),
                            minute=int(time_split[1]),
                            second=int(time_split[2]),
                        )
                        if time is None or dtime > time:
                            time = dtime
                        jobcount += 1
        return time, jobcount

    @property
    def current_jobs(self):
        running_jobs = []
        for node in self._job_nodes:
            if (
                node.environment["start_time_stamp"] is not None
                and node.environment["status"] is None
            ):
                running_jobs.append(node)
        if len(running_jobs) == 0:
            return None
        else:
            return running_jobs


class DummyLock:
    def __init__(self):
        self.obtained = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        raise NotImplementedError(
            "A dummy pipeline lock is being used. An actual pipeline lock should be implemented"
        )



from __future__ import annotations

from pathlib import Path

from pipeliner.api.api_utils import (
    edit_jobstar,
    job_parameters_dict,
    write_default_jobstar,
)
from pipeliner.api.manage_project import PipelinerProject

cluster_options = {
    "do_queue": "Yes",
    "qsubscript": "/dls_sw/apps/EM/relion/qsub_templates/qsub_template_hamilton_pipeliner",
    "use_gpu": "Yes",
    "gpu_ids": "0:1:2:3",
    "nr_mpi": 5,
    "nr_threads": 8,
}


class RefinePipelineRunner:
    def __init__(
        self,
        project_path: str,
        particles_star_file: str,
        ref_model: str,
        mask: str = "",
        particle_diameter: float = 170,
        autob_highres: float = 4.75,
        extract_size: int = 0,
        symmetry: str = "C1",
        ini_high: float = 60,
    ):
        self._proj_path = Path(project_path)
        self._particles_star = Path(particles_star_file)
        self._proj = PipelinerProject()
        self._ref_model = ref_model
        self._default_params = {
            "relion.refine3d": {
                "particle_diameter": particle_diameter,
                "sym_name": symmetry,
                "ini_high": ini_high,
            },
            "relion.postprocess": {"other_args": f"--autob_highres {autob_highres}"},
        }
        self._mask = mask
        self._extract_size = extract_size

    def _run_job(self, job: str, params: dict, cluster=True) -> str:
        write_default_jobstar(job)
        _params = job_parameters_dict(job)
        _params.update(params)
        _params.update(self._default_params.get(job, {}))
        if cluster:
            _params.update(cluster_options)
        edit_jobstar(
            f"{job.replace('.', '_')}_job.star",
            _params,
            f"{job.replace('.', '_')}_job.star",
        )
        job_path = self._proj.run_job(
            f"{job.replace('.', '_')}_job.star", wait_for_queued=True
        )
        return job_path

    def _run_import(
        self, fn_in: str = "", node_type: str = "Particles STAR file (.star)"
    ) -> str:
        job = "relion.import.other"
        params = {
            "fn_in_other": fn_in or str(self._particles_star),
            "node_type": node_type,
        }
        return self._run_job(job, params, cluster=False)

    def _run_refine3d(self, imported_star_file: str):
        job = "relion.refine3d"
        if "job" in self._ref_model:
            model = self._ref_model
        else:
            model_import = self._run_import(
                fn_in=self._ref_model, node_type="3D reference (.mrc)"
            )
            model = f"{model_import}/{Path(self._ref_model).name}"
        params = {"fn_img": imported_star_file, "fn_ref": model}
        return self._run_job(job, params)

    def _run_postprocess(self, input_model: str):
        job = "relion.postprocess"
        if "job" in self._mask:
            mask = self._mask
        else:
            mask_import = self._run_import(
                fn_in=self._mask, node_type="3D reference (.mrc)"
            )
            mask = f"{mask_import}/{Path(self._mask).name}"
        params = {"fn_in": input_model, "fn_mask": mask, "angpix": -1}
        return self._run_job(job, params, cluster=False)

    def _run_extract(
        self,
        particles_star: str,
        micrographs: str = "CtfFind/job003/micrographs_ctf.star",
    ):
        job = "relion.extract.reextract"
        params = {
            "star_mics": micrographs,
            "fndata_reextract": particles_star,
            "extract_size": self._extract_size,
        }
        return self._run_job(job, params)

    def __call__(self, micrographs_star: str = "CtfFind/job003/micrographs_ctf.star"):
        if "/job" in str(self._particles_star):
            import_path = str(self._particles_star)
        else:
            import_path = self._run_import() + f"/{self._particles_star.name}"
        if self._extract_size:
            import_path = (
                self._run_extract(import_path, micrographs=micrographs_star)
                + "/particles.star"
            )
        refine_path = self._run_refine3d(import_path)
        half_map = list(Path(refine_path).glob("run_half1*unfil.mrc"))[0]
        self._run_postprocess(str(half_map.relative_to(self._proj_path)))



#!/usr/bin/env python
"""
External job for calling cryolo within Relion 3.1
in_mics are the micrographs to be picked on
in_model is the model to be used, empty will use general model!

Run in main Relion project directory
external_cryolo.py --o $PATH_WHERE_TO_STORE --in_mics $PATH_TO_MCORR/CTF_STARFILE --box_size $BOX_SIZE --threshold 0.3 (optional: --in_model $PATH_TO_MODEL)
eg:
cryolo_external_job.py --o "External/crYOLO_AutoPick" --in_mics "CtfFind/job004/micrographs_ctf.star" --box_size 300 --threshold 0.3
"""

from __future__ import annotations

import argparse
import datetime
import json
import os
import os.path
import pathlib
import platform
import shutil
import subprocess
import sys

import gemmi

RELION_JOB_FAILURE_FILENAME = "RELION_JOB_EXIT_FAILURE"
RELION_JOB_SUCCESS_FILENAME = "RELION_JOB_EXIT_SUCCESS"


def run_job(project_dir, job_dir, args_list):
    parser = argparse.ArgumentParser()
    parser.add_argument("--in_mics", help="Input micrographs STAR file")
    parser.add_argument(
        "--j", dest="threads", help="Number of threads to run (ignored)"
    )
    parser.add_argument("--box_size", help="Size of box (~ particle size)")
    parser.add_argument("--threshold", help="Threshold for picking (default = 0.3)")
    parser.add_argument("--in_model", help="model from previous job")
    parser.add_argument("--gmodel", help="cryolo general model")
    parser.add_argument("--config", help="cryolo config")
    parser.add_argument("--gpu", help='GPUs to use (e.g. "0 1 2 3")')
    parser.add_argument(
        "--from_movies",
        type=int,
        help="Whether or not initial images were movies",
        default=1,
    )
    args = parser.parse_args(args_list)
    thresh = args.threshold
    box_size = args.box_size
    gen_model = args.gmodel
    conf_file = args.config
    gpus = args.gpu

    # Use general model by default if in_model not given or doesn't exist
    model = gen_model
    if args.in_model is not None:
        model_path = os.path.join(project_dir, args.in_model)
        if os.path.isfile(model_path):
            model = model_path
        else:
            print(f" cryolo_external_job: Cannot find fine tuned model {model_path}")

    # Making a cryolo config file with the correct box size
    with open(conf_file, "r") as json_file:
        data = json.load(json_file)
        data["model"]["anchors"] = [int(box_size), int(box_size)]
    with open("config.json", "w") as outfile:
        json.dump(data, outfile)

    # Reading the micrographs star file from relion
    in_doc = gemmi.cif.read_file(os.path.join(project_dir, args.in_mics))
    data_as_dict = json.loads(in_doc.as_json())["micrographs"]

    try:
        os.mkdir("cryolo_input")
    except FileExistsError:
        # Not crucial so if fails due to any reason just carry on
        try:
            with open(
                "done_mics.txt", "a+"
            ) as f:  # Done mics is to ensure that cryolo doesn't pick from already done mics
                for micrograph in os.listdir("cryolo_input"):
                    f.write(micrograph + "\n")
        except Exception:
            pass
        shutil.rmtree("cryolo_input")
        os.mkdir("cryolo_input")

    # Arranging files for cryolo to predict particles from
    # Not crucial so if fails due to any reason just carry on
    try:
        with open("done_mics.txt", "r") as f:
            done_mics = f.read().splitlines()
    except Exception:
        done_mics = []

    # Count new micrographs because cryolo fails if its input directory is empty
    mic_count = 0
    for micrograph in data_as_dict["_rlnmicrographname"]:
        if os.path.split(micrograph)[-1] not in done_mics:
            os.link(
                os.path.join(project_dir, micrograph),
                os.path.join(
                    project_dir, job_dir, "cryolo_input", os.path.split(micrograph)[-1]
                ),
            )
            mic_count += 1

    if mic_count > 0:
        print(" cryolo_external_job: Running from model {}".format(model))
        cryolo_command = [
            "cryolo_predict.py",
            "--conf",
            "config.json",
            "-i",
            f"{pathlib.PurePosixPath(project_dir) / job_dir / 'cryolo_input'}",
            "-o",
            f"{pathlib.PurePosixPath(project_dir) / job_dir / 'gen_pick'}",
            "--weights",
            model,
            "--gpu",
            gpus,
            "--threshold",
            thresh,
        ]
        subprocess.run(cryolo_command, check=True)
    else:
        print(" cryolo_external_job: No new micrographs, not running crYOLO")

    os.makedirs("picked_stars", exist_ok=True)

    # Arranging files for Relion to use
    for picked in os.listdir(os.path.join(project_dir, job_dir, "gen_pick", "STAR")):
        new_name = os.path.splitext(picked)[0] + "_autopick" + ".star"
        try:
            os.link(
                os.path.join(project_dir, job_dir, "gen_pick", "STAR", picked),
                os.path.join(project_dir, job_dir, "picked_stars", new_name),
            )
        except FileExistsError:
            pass

    # Writing a star file for Relion
    with open("coords_suffix_autopick.star", "w") as part_file:
        part_file.write(args.in_mics)

    # Required star file
    out_doc = gemmi.cif.Document()
    output_nodes_block = out_doc.add_new_block("output_nodes")
    loop = output_nodes_block.init_loop(
        "", ["_rlnPipeLineNodeName", "_rlnPipeLineNodeType"]
    )
    loop.add_row([os.path.join(job_dir, "coords_suffix_autopick.star"), "2"])
    out_doc.write_file("RELION_OUTPUT_NODES.star")
    ctf_star = os.path.join(project_dir, args.in_mics)
    path_drop = 2 if args.from_movies else 0
    correct_paths(ctf_star, path_drop=path_drop)


def correct_paths(ctf_star, path_drop: int = 2):
    in_doc = gemmi.cif.read_file(ctf_star)
    data_as_dict = json.loads(in_doc.as_json())["micrographs"]

    for i in range(len(data_as_dict["_rlnmicrographname"])):
        name = data_as_dict["_rlnmicrographname"][i]
        dirs, mic_file = os.path.split(name)
        full_dir = ""
        for d in dirs.split("/")[path_drop:]:
            full_dir = os.path.join(full_dir, d)
        os.makedirs(full_dir, exist_ok=True)
        picked_star = os.path.splitext(mic_file)[0] + "_autopick.star"
        try:
            shutil.move(
                os.path.join("picked_stars", picked_star),
                os.path.join(full_dir, picked_star),
            )
        except FileNotFoundError:
            print(f" cryolo_external_job: cryolo found no particles in {picked_star}")


def main():
    """Change to the job working directory, then call run_job()"""
    parser = argparse.ArgumentParser()
    parser.add_argument("--o", dest="out_dir", help="Output directory name")
    parser.add_argument(
        "--pipeline_control", help="Directory for pipeline control files"
    )
    known_args, other_args = parser.parse_known_args()
    project_dir = os.getcwd()
    job_dir = known_args.out_dir

    # pipeline_control_dir is probably the same as job_dir, and we should use job_dir
    # if pipeline_control_dir is not given. Make the path absolute (by joining it onto
    # project_dir) so success and failure indicator files are always written in the
    # right place even if the working directory is changed.
    pipeline_control_dir = (
        os.path.join(project_dir, known_args.pipeline_control)
        if known_args.pipeline_control is not None
        else os.path.join(project_dir, job_dir)
    )
    # Ensure job_dir and pipeline_control_dir (probably the same) both exist
    os.makedirs(job_dir, exist_ok=True)
    os.makedirs(pipeline_control_dir, exist_ok=True)

    # Prepare indicator filenames for use after the job has finished
    failure_filename = os.path.join(pipeline_control_dir, RELION_JOB_FAILURE_FILENAME)
    success_filename = os.path.join(pipeline_control_dir, RELION_JOB_SUCCESS_FILENAME)

    # Print current time and host to stdout and stderr
    # This helps with debugging when the job is run repeatedly
    time = datetime.datetime.now()
    host = platform.node()
    message = f"{time}: cryolo_external_job.py running on {host}"
    print(message)
    print(message, file=sys.stderr)

    # Change to the job directory to actually run the job
    os.chdir(job_dir)
    try:
        run_job(project_dir, job_dir, other_args)
    except Exception:
        open(failure_filename, "w").close()
        raise
    else:
        open(success_filename, "w").close()


if __name__ == "__main__":
    main()



from __future__ import annotations

import collections.abc

from relion.dbmodel.dbgraph import DBGraph
from relion.dbmodel.dbnode import DBNode
from relion.dbmodel.modeltables import (
    CryoemInitialModelTable,
    CTFTable,
    MotionCorrectionTable,
    ParticleClassificationGroupTable,
    ParticleClassificationTable,
    ParticlePickerTable,
    RelativeIceThicknessTable,
)


class DBModel(collections.abc.Mapping):
    def __init__(self, db_name):
        self._db = self._make_db(db_name)

    def __eq__(self, other):
        if isinstance(other, DBModel):
            if super().__eq__(other) and self._db == other._db:
                return True
        return False

    def __getitem__(self, key):
        return self._db[key]

    def __len__(self):
        return len(self._db.keys())

    def __iter__(self):
        return self._db.items()

    def values(self):
        return self._db.values()

    def keys(self):
        return self._db.keys()

    @property
    def db_nodes(self):
        nodes = []
        for node in self.values():
            if node not in nodes:
                nodes.append(node)
        return nodes

    def _make_db(self, db_name):
        if db_name == "ISPyB":
            return self._make_ispyb_model()
        else:
            raise ValueError(f"{db_name} not implmented in relion.dbmodel.DBModel")

    def _make_ispyb_model(self):
        self.mc_db_node = DBNode("MCTable", [MotionCorrectionTable()])

        self.ctf_db_node = DBNode("CTFTable", [CTFTable()])
        self.mc_db_node.link_to(
            self.ctf_db_node,
            traffic={
                "check_for": "micrograph_full_path",
                "foreign_key": "motion_correction_id",
                "table_key": "motion_correction_id",
                "foreign_table": self.mc_db_node.tables[0],
            },
        )

        self.rel_ice_db_node = DBNode(
            "RelativeIceThicknessTable", [RelativeIceThicknessTable()]
        )
        self.mc_db_node.link_to(
            self.rel_ice_db_node,
            traffic={
                "check_for": "micrograph_full_path",
                "foreign_key": "motion_correction_id",
                "table_key": "motion_correction_id",
                "foreign_table": self.mc_db_node.tables[0],
            },
        )

        self.parpick_db_node = DBNode("ParticlePickerTable", [ParticlePickerTable()])
        self.mc_db_node.link_to(
            self.parpick_db_node,
            traffic={
                "check_for": "micrograph_full_path",
                "foreign_key": "motion_correction_id",
                "table_key": "first_motion_correction_id",
                "foreign_table": self.mc_db_node.tables[0],
            },
        )

        self.class_group_db_node = DBNode(
            "ClassificationGroupTable",
            [ParticleClassificationGroupTable()],
        )
        self.class_db_node = DBNode(
            "ClassificationTable",
            [ParticleClassificationTable()],
        )
        self.parpick_db_node.link_to(
            self.class_group_db_node,
            traffic={
                "check_for": "parpick_job_string",
                "check_for_foreign_name": "job_string",
                "foreign_key": "particle_picker_id",
                "table_key": "particle_picker_id",
                "first": True,
                "foreign_table": self.parpick_db_node.tables[0],
            },
        )
        self.class_group_db_node.link_to(
            self.class_db_node,
            traffic={
                "check_for": "job_string",
                "foreign_key": "particle_classification_group_id",
                "table_key": "particle_classification_group_id",
                "foreign_table": self.class_group_db_node.tables[0],
            },
        )
        self.class2d_db_node = DBGraph(
            "2DClassificationTables",
            [self.class_group_db_node, self.class_db_node],
            auto_connect=True,
        )

        self.ini_model_db_node = DBNode(
            "InitialModelTable",
            [CryoemInitialModelTable()],
        )
        self.class_db_node.link_to(
            self.ini_model_db_node,
            traffic={
                "check_for": "job_string",
                "foreign_key": "particle_classification_id",
                "table_key": "particle_classification_id",
                "foreign_table": self.class_db_node.tables[0],
                "first": False,
            },
        )
        self.class3d_db_node = DBGraph(
            "3DClassificationTables",
            [
                self.class_group_db_node,
                self.class_db_node,
                self.ini_model_db_node,
            ],
            auto_connect=True,
        )

        db_dict = {
            "MotionCorr": self.mc_db_node,
            "CtfFind": self.ctf_db_node,
            "AutoPick": self.parpick_db_node,
            "External/crYOLO_AutoPick/": self.parpick_db_node,
            "Class2D": self.class2d_db_node,
            "InitialModel": self.class3d_db_node,
            "Class3D": self.class3d_db_node,
            "External/Icebreaker_5fig/": self.rel_ice_db_node,
            "IceBreaker/Icebreaker_5fig/": self.rel_ice_db_node,
        }

        return db_dict



from __future__ import annotations

import argparse
import pathlib

from relion import Project


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("proj_path")
    args = parser.parse_args()
    relion_dir = pathlib.Path(args.proj_path)
    proj = Project(relion_dir)
    current_jobs = proj.current_jobs
    if current_jobs is None:
        print("There are not any currently running jobs")
    else:
        print("These jobs are currently running: \n")
        for current_job in current_jobs:
            alias = current_job.environment["alias"]
            if alias is not None:
                print(f"Current job: {current_job._path} [alias={alias}]")
            else:
                print(f"Current job: {current_job._path}")
            print(f"Job started at: {current_job.environment['start_time_stamp']}")
            print(f"Job has been run {current_job.environment['job_count']} time(s)")
            print()


if __name__ == "__main__":
    run()



from __future__ import annotations

import argparse

from relion.pipeline.refine import RefinePipelineRunner


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--dir",
        help="Working directory where Relion will run",
        dest="working_directory",
        default=".",
    )
    parser.add_argument(
        "-f",
        "--particles-file",
        help="Star file containing particles to be imported",
        action="append",
        dest="particles_files",
    )
    parser.add_argument(
        "-m",
        "--ref-model",
        help="Reference model",
        dest="ref_model",
    )
    parser.add_argument(
        "--mask",
        help="Path to mask for post processing job",
        dest="mask",
        default="",
    )
    parser.add_argument(
        "--particle-diameter",
        help="Particle diameter in Angstroms",
        dest="particle_diameter",
        default=170,
    )
    parser.add_argument(
        "--extract-size",
        help="Extraction box size",
        dest="extract_size",
        default=256,
    )
    parser.add_argument(
        "--micrographs-star",
        help="Path to micrographs star file",
        dest="micrographs_star",
        default="CtfFind/job003/micrographs_ctf.star",
    )
    parser.add_argument(
        "--autob-highres",
        help="High resolution limit for b-factor calculation",
        dest="autob_highres",
        default=4.75,
    )
    parser.add_argument(
        "--ini-high",
        help="Low pass filter for 3D refinement",
        dest="ini_high",
        default=60,
    )
    parser.add_argument(
        "--sym",
        help="Symmetry for 3D refinement",
        dest="sym",
        default="C1",
    )
    args = parser.parse_args()

    for pf in args.particles_files:
        runner = RefinePipelineRunner(
            args.working_directory,
            pf,
            args.ref_model,
            mask=args.mask,
            autob_highres=args.autob_highres,
            extract_size=args.extract_size,
            symmetry=args.sym,
            ini_high=args.ini_high,
        )
        runner(micrographs_star=args.micrographs_star)



from __future__ import annotations

from pathlib import Path

import procrunner
import workflows.recipe
from pydantic import BaseModel, Field
from pydantic.error_wrappers import ValidationError
from workflows.services.common_service import CommonService


class CTFParameters(BaseModel):
    pix_size: float
    voltage: float = 300.0
    spher_aber: float = 2.7
    ampl_contrast: float = 0.8
    ampl_spectrum: int = 512
    min_res: float = 30.0
    max_res: float = 5.0
    min_defocus: float = 5000.0
    max_defocus: float = 50000.0
    defocus_step: float = 100.0
    astigmatism_known: str = "no"
    slow_search: str = "no"
    astigmatism_restrain: str = "no"
    additional_phase_shift: str = "no"
    expert_options: str = "no"
    input_image: str = Field(..., min_length=1)
    output_image: str = Field(..., min_length=1)
    mc_uuid: int


class CTFFind(CommonService):
    """
    A service for CTF estimating micrographs with CTFFind
    """

    # Human readable service name
    _service_name = "DLS CTFFind"

    # Logger name
    _logger_name = "relion.zocalo.ctffind"

    # Values to extract for ISPyB
    astigmatism_angle: float
    cc_value: float
    estimated_resolution: float
    defocus1: float
    defocus2: float

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("CTFFind service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "ctffind",
            self.ctf_find,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_ctf_output(self, line: str):
        if not line:
            return

        try:
            if line.startswith("Estimated defocus values"):
                line_split = line.split()
                self.defocus1 = float(line_split[4])
                self.defocus2 = float(line_split[6])
            if line.startswith("Estimated azimuth"):
                line_split = line.split()
                self.astigmatism_angle = float(line_split[4])
            if line.startswith("Score"):
                line_split = line.split()
                self.cc_value = float(line_split[2])
            if line.startswith("Thon rings"):
                line_split = line.split()
                self.estimated_resolution = float(line_split[8])
        except Exception as e:
            self.log.warning(f"{e}")

    def ctf_find(self, rw, header: dict, message: dict):
        class RW_mock:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            print(
                "Incoming message is not a recipe message. Simple messages can be valid"
            )
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return
            self.log.debug("Received a simple message")

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = RW_mock()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        command = ["ctffind"]

        try:
            if isinstance(message, dict):
                ctf_params = CTFParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                ctf_params = CTFParameters(**{**rw.recipe_step.get("parameters", {})})
        except (ValidationError, TypeError):
            self.log.warning(
                f"CTF estimation parameter validation failed for message: {message} and recipe parameters: {rw.recipe_step.get('parameters', {})}"
            )
            rw.transport.nack(header)
            return

        parameters_list = [
            ctf_params.input_image,
            ctf_params.output_image,
            ctf_params.pix_size,
            ctf_params.voltage,
            ctf_params.spher_aber,
            ctf_params.ampl_contrast,
            ctf_params.ampl_spectrum,
            ctf_params.min_res,
            ctf_params.max_res,
            ctf_params.min_defocus,
            ctf_params.max_defocus,
            ctf_params.defocus_step,
            ctf_params.astigmatism_known,
            ctf_params.slow_search,
            ctf_params.astigmatism_restrain,
            ctf_params.additional_phase_shift,
            ctf_params.expert_options,
        ]

        parameters_string = "\n".join(map(str, parameters_list))
        self.log.info(
            f"Input: {ctf_params.input_image} Output: {ctf_params.output_image}"
        )
        result = procrunner.run(
            command=command,
            stdin=parameters_string.encode("ascii"),
            callback_stdout=self.parse_ctf_output,
        )
        if result.returncode:
            self.log.error(
                f"CTFFind failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        # Extract results for ispyb
        astigmatism = self.defocus2 - self.defocus1
        estimated_defocus = (self.defocus1 + self.defocus2) / 2

        ispyb_parameters = {
            "box_size_x": str(ctf_params.ampl_spectrum),
            "box_size_y": str(ctf_params.ampl_spectrum),
            "min_resolution": str(ctf_params.min_res),
            "max_resolution": str(ctf_params.max_res),
            "min_defocus": str(ctf_params.min_defocus),
            "max_defocus": str(ctf_params.max_defocus),
            "astigmatism": str(astigmatism),
            "defocus_step_size": str(ctf_params.defocus_step),
            "astigmatism_angle": str(self.astigmatism_angle),
            "estimated_resolution": str(self.estimated_resolution),
            "estimated_defocus": str(estimated_defocus),
            "amplitude_contrast": str(ctf_params.ampl_contrast),
            "cc_value": str(self.cc_value),
            "fft_theoretical_full_path": str(
                Path(ctf_params.output_image).with_suffix(".jpeg")
            ),  # path to output mrc (would be jpeg if we could convert in SW)
        }

        # Forward results to ispyb
        ispyb_parameters.update(
            {
                "ispyb_command": "buffer",
                "buffer_lookup": {"motion_correction_id": ctf_params.mc_uuid},
                "buffer_command": {"ispyb_command": "insert_ctf"},
            }
        )
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, RW_mock):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb", ispyb_parameters)

        # Forward results to images service
        self.log.info(f"Sending to images service {ctf_params.output_image}")
        if isinstance(rw, RW_mock):
            rw.transport.send(
                destination="images",
                message={
                    "parameters": {"images_command": "mrc_to_jpeg"},
                    "file": ctf_params.output_image,
                },
            )
        else:
            rw.send_to(
                "images",
                {
                    "parameters": {"images_command": "mrc_to_jpeg"},
                    "file": ctf_params.output_image,
                },
            )

        rw.transport.ack(header)



from __future__ import annotations

from relion.node import Node

try:
    from graphviz import Digraph
except ImportError:
    pass


class Graph(Node):
    def __init__(self, name, node_list, auto_connect=False):
        super().__init__(name)
        self._node_list = node_list
        try:
            self.origins = self.find_origins()
        except IndexError:
            self.origins = []
        self._call_returns = {}
        self._called_nodes = []
        self._traversed = []
        if auto_connect:
            self._check_connections()

    def __eq__(self, other):
        if isinstance(other, Graph):
            if len(self.nodes) == len(other.nodes):
                for n in self.nodes:
                    if n not in other.nodes:
                        return False
                return True
        return False

    def __hash__(self):
        return hash(
            ("relion.node.graph.Graph", self._name, iter(self._node_list), self.nodeid)
        )

    def __len__(self):
        return len(self._node_list)

    def __getitem__(self, index):
        if not isinstance(index, int):
            raise ValueError("Index of Graph must be an integer")
        return self._node_list[index]

    def __call__(self, *args, **kwargs):
        for node in self._node_list:
            node.environment.set_escalate(self.environment)
        res = super().__call__(*args, **kwargs)
        return res

    @property
    def nodes(self):
        return self._node_list

    def func(self, *args, **kwargs):
        self._call_returns = {}
        if self._in_multi_call:
            for node in self.origins:
                node._completed = self._completed
        self.traverse()
        self._traversed = []
        self._called_nodes = []
        for node in self.nodes:
            node.environment.reset()
            node._completed = []
        if self._call_returns == {}:
            return None
        else:
            return self._call_returns

    def _check_connections(self):
        for node in self._node_list:
            for i_node in node._in:
                if i_node not in self._node_list:
                    i_node.link_to(self)
                    i_node._link_traffic[self.nodeid].update(
                        i_node._link_traffic[node.nodeid]
                    )

    def extend(self, other):
        if not isinstance(other, Graph):
            raise ValueError("Can only extend a ProtoGraph with another Graph")
        self._node_list.extend(other._node_list)

    def index(self, node):
        return self._node_list.index(node)

    def link_from_to(self, from_node, to_node):
        self[self.index(from_node)].link_to(to_node)

    def node_explore(self, node, explored):
        if not isinstance(node, Node):
            raise ValueError(
                "Graph.node_explore must be called with a ProtoNode as the starting point; a string or similar is insufficient"
            )
        if node not in explored:
            explored.append(node)
        for next_node in node:
            self.node_explore(next_node, explored)

    def add_node(self, new_node, auto_connect=False):
        if isinstance(new_node, Node):
            self._node_list.append(new_node)
            if auto_connect:
                for i_node in new_node._in:
                    if i_node not in self._node_list:
                        i_node.link_to(self)
            new_node.environment.set_escalate(self.environment)
        else:
            raise ValueError("Attempted to add a node that was not a Node")

    def remove_node(self, node_name, advance=False):
        behind_nodes = []
        for currnode in self._node_list:
            if currnode.name == str(node_name):
                if currnode.environment.propagate.released:
                    for next_node in currnode:
                        next_node.environment.update_prop(
                            currnode.environment.propagate.store
                        )
                        if advance:
                            next_node.environment.update(
                                currnode.environment.propagate.store
                            )
            if node_name in currnode:
                behind_nodes.append(currnode)
                currnode.unlink_from(node_name)
            if node_name in currnode._in:
                currnode._in.remove(node_name)
        for bnode in behind_nodes:
            for next_node in self._node_list[self._node_list.index(node_name)]:
                bnode.link_to(next_node)
        self._node_list.remove(node_name)

    def find_origins(self):
        child_nodes = []
        for node in self.nodes:
            child_nodes.extend(node)
        origins = [p for p in self.nodes if p not in child_nodes]
        return origins

    def merge(self, other):
        node_names = [p._name for p in self.nodes]
        other_names = [p._name for p in other.nodes]
        if len(set(node_names).intersection(set(other_names))) > 0:
            for new_node in other.nodes:
                if new_node not in self.nodes:
                    self.add_node(new_node)
                else:
                    for next_node in new_node:
                        if next_node not in self.nodes[self.index(new_node)]:
                            self.nodes[self.index(new_node)].link_to(next_node)
            return True
        else:
            return False

    def traverse(self):
        for o in self.origins:
            self._follow(o, traffic={}, share=[], append=o._can_append)

    def _follow(self, node, traffic, share, append=False):
        called = False
        if node not in self.nodes:
            return
        if node.nodeid in self._called_nodes:
            called = True
        node.environment.update(traffic, can_append_list=append)

        for sh in share:
            node.environment[sh[1]] = sh[0]

        if (
            all(n in node._completed for n in node._in)
            and node.nodeid not in self._called_nodes
        ):
            called = True

            self._call_returns[node.nodeid] = node()
            self._called_nodes.append(node.nodeid)

        for next_node in node:
            next_node.environment.update_prop(node.environment.propagate)
            next_traffic = node._link_traffic.get(next_node.nodeid, {})
            if next_traffic is None:
                next_traffic = self._call_returns[node.nodeid]
            next_share = []
            if node._share_traffic.get(next_node.nodeid) is not None:
                for sh in node._share_traffic[next_node.nodeid]:
                    next_share.append((node.environment[sh[0]], sh[1]))
            if (node.nodeid, next_node.nodeid) not in self._traversed and called:
                self._traversed.append((node.nodeid, next_node.nodeid))
                self._follow(
                    next_node,
                    next_traffic,
                    next_share,
                    append=node._can_append,
                )
            elif not called:
                self._follow(
                    next_node,
                    next_traffic,
                    next_share,
                    append=node._can_append,
                )

    def show(self):
        try:
            digraph = Digraph(format="svg", strict=True)
            digraph.attr(rankdir="LR", splines="ortho")
            for node in self._node_list:
                if isinstance(node, Graph):
                    digraph.node(name=str(node.name), shape="box")
                    for gnode in node._node_list:
                        if isinstance(gnode, Graph):
                            digraph.node(name=str(gnode.name), shape="box")
                        else:
                            digraph.node(name=str(gnode.name), shape=gnode.shape)
                        digraph.edge(str(node.name), str(gnode.name), style="dashed")
                        for next_gnode in gnode:
                            digraph.edge(str(gnode.name), str(next_gnode.name))
                else:
                    digraph.node(name=str(node.name), shape=node.shape)
                for next_node in node:
                    if next_node in self._node_list:
                        digraph.edge(str(node.name), str(next_node.name))
            digraph.render(f"{self.name}.gv")
        except Exception:
            raise Warning(
                "Failed to create nodes display. Your environment may not have graphviz available."
            )
